require 'pathname'

module Caligari
  # This module contains constants representing XDG configuration directories
  # and files for Caligari as specified by the XDG Base Directory
  # Specification.
  #
  # @see http://polr.me/zjp XDG Base Directory Specification
  module XDG
    # Collects configuration directories as defined by the XDG Base Directory
    # Specification as `Pathname` instances.
    #
    # @api private
    # @return [Array<Pathname>] the collected XDG configuration directories
    def self.collect_xdg_config_dirs
      dirs = ENV['XDG_CONFIG_DIRS']
      if dirs
        dirs.split(':').map! do |dir|
          Pathname.new(dir).expand_path.freeze
        end.freeze
      else
        [Pathname.new('/etc/xdg').freeze].freeze
      end
    end
    private_class_method :collect_xdg_config_dirs

    # The global XDG configuration directories as an array of `Pathname`
    # instances.
    CONFIG_DIRS = collect_xdg_config_dirs

    # The user XDG configuration directory as a `Pathname` instance.
    CONFIG_HOME =
      Pathname.new(ENV['XDG_CONFIG_HOME'] || '~/.config').expand_path.freeze

    # The existent XDG configuration files for Caligari as a `Pathname`
    # instance.
    CONFIG_FILES = (CONFIG_DIRS + [CONFIG_HOME]).map! do |dir|
      (dir + 'caligari.yml').freeze
    end.select! { |file| file.exist? }.freeze
  end
end
